#include "host.h"

#include "immap_83xx.h"

#define PCIEXP(d) ((pex83xx_t *)(d->pciexp))

static u32 fix_reg_bits(struct devdata *devdata)
{
	pex83xx_t *pciexp = PCIEXP(devdata);
	u32 reg_bits;

	writel(0, &pciexp->bridge.pex_ombcr);

	reg_bits = readl(&pciexp->bridge.pex_ombdr);

	PRINTK_DEBUG("pex_ombdr: %08x", reg_bits);

	if (reg_bits) {
		writel(0, &pciexp->bridge.pex_ombdr);

		devdata->reg_bits |= reg_bits;
	}

	return reg_bits;
}

static void sgs_reset_obw(struct devdata *devdata)
{
	pex83xx_t *pciexp = PCIEXP(devdata);

	// setup outbound window translation address
	writel(0, &pciexp->bridge.pex_outbound_win[0].tarl);
	writel(0, &pciexp->bridge.pex_outbound_win[0].tarh);

	writel(0, &pciexp->bridge.pex_outbound_win[1].tarl);
	writel(0, &pciexp->bridge.pex_outbound_win[1].tarh);

	writel(0, &pciexp->bridge.pex_outbound_win[2].tarl);
	writel(0, &pciexp->bridge.pex_outbound_win[2].tarh);

	writel(0, &pciexp->bridge.pex_outbound_win[3].tarl);
	writel(0, &pciexp->bridge.pex_outbound_win[3].tarh);
}

static void sgs_setup_old(struct devdata *devdata)
{
	pex83xx_t *pciexp = PCIEXP(devdata);

	devdata->host2dev = &devdata->iomem->old.host2dev;
	devdata->host_read_wait = &devdata->iomem->old.host_read_wait;
	devdata->host_write_wait = &devdata->iomem->old.host_write_wait;
	devdata->dev2host = &devdata->iomem->old.dev2host;
	devdata->dev_read_wait = &devdata->iomem->old.dev_read_wait;
	devdata->dev_write_wait = &devdata->iomem->old.dev_write_wait;

	sgs_reset_obw(devdata);

	fix_reg_bits(devdata);

	// PCI Express Host Interrupt Enable Register
	// OPCIE (0x1): Outbound PIO transaction completed
	// OPAIE (0x2): Outbound PIO transaction abort
	// IPCIE (0x4): Inbound PIO transaction completed
	// IPAIE (0x8): Inbound PIO transaction aborted
	// WDCIE (0x10): WDMA chain descriptor transfer completed
	// WDSIE (0x20): WDMA descriptor transfer completed
	// WDAIE (0x40): WDMA transfer aborted
	// RDCIE (0x80): RDMA chain descriptor transfer completed
	// RDSIE (0x100): RDMA descriptor transfer completed
	// RDAIE (0x200): RDMA transfer aborted
	// OMBIE (0x400): Outbound mailbox ready
	// CBRIE (0x800): CSB bridge reset
	writel(0x00000400, &pciexp->bridge.pex_hier);

	// get EP physcal address of out- and inbound window (needed for DMA)
	devdata->ib_phys_addr = readl(&pciexp->bridge.pex_epiwtar0) & ~4095;
}

static void sgs_reset_old(struct devdata *devdata)
{
	pex83xx_t *pciexp = PCIEXP(devdata);

	writel(0x00000000, &pciexp->bridge.pex_hier);

	sgs_reset_obw(devdata);
}

static void sgs_setup(struct devdata *devdata)
{
	pex83xx_t *pciexp = PCIEXP(devdata);

	devdata->host2dev = &devdata->iomem->new.host2dev.fifo;
	devdata->host_read_wait = &devdata->iomem->new.dev2host.read_wait;
	devdata->host_write_wait = &devdata->iomem->new.host2dev.write_wait;
	devdata->dev2host = &devdata->iomem->new.dev2host.fifo;
	devdata->dev_read_wait = &devdata->iomem->new.host2dev.read_wait;
	devdata->dev_write_wait = &devdata->iomem->new.dev2host.write_wait;

	sgs_reset_obw(devdata);

	fix_reg_bits(devdata);

	writel(0x00000400, &pciexp->bridge.pex_hier);

	// get EP physcal address of out- and inbound window (needed for DMA)
	devdata->ib_phys_addr = readl(&pciexp->bridge.pex_epiwtar0) & ~4095;
}

static void sgs_reset(struct devdata *devdata)
{
	pex83xx_t *pciexp = PCIEXP(devdata);

	writel(0x00000000, &pciexp->bridge.pex_hier);

	sgs_reset_obw(devdata);
}

static void sgs_link(struct devdata *devdata, struct devdata *linkdev)
{
	pex83xx_t *pciexp = PCIEXP(devdata);

	// setup outbound window translation address
	writel(LO32(linkdev->iobar), &pciexp->bridge.pex_outbound_win[0].tarl);
	writel(HI32(linkdev->iobar), &pciexp->bridge.pex_outbound_win[0].tarh);
	writel(LO32(linkdev->pciexbar), &pciexp->bridge.pex_outbound_win[1].tarl);
	writel(HI32(linkdev->pciexbar), &pciexp->bridge.pex_outbound_win[1].tarh);
	writel(LO32(linkdev->fpgabar), &pciexp->bridge.pex_outbound_win[2].tarl);
	writel(HI32(linkdev->fpgabar), &pciexp->bridge.pex_outbound_win[2].tarh);
}

static void sgs_unlink(struct devdata *devdata)
{
	sgs_reset_obw(devdata);
}

static void sgs_gen_inbound_mbox_irq(struct devdata * devdata)
{
	pex83xx_t *pciexp = PCIEXP(devdata);

	writel(1, &pciexp[0].bridge.pex_imbcr);
}

static irqreturn_t _sgs_isr(struct devdata * devdata)
{
	pex83xx_t *pciexp = PCIEXP(devdata);
	u32 hisr;
	u32 hisr_handled = 0;

	PRINTK_DEBUG("%s:%d\n", __FUNCTION__, __LINE__);


	for(;;) {
		// PCI Express Host Interrupt Status Register
		// OPCI (0x1): Outbound PIO transaction completed
		// OPAI (0x2): Outbound PIO transaction abort
		// IPCI (0x4): Inbound PIO transaction completed
		// IPAI (0x8): Inbound PIO transaction aborted
		// WDCI (0x10): WDMA chain descriptor transfer completed
		// WDSI (0x20): WDMA descriptor transfer completed
		// WDAI (0x40): WDMA transfer aborted
		// RDCI (0x80): RDMA chain descriptor transfer completed
		// RDSI (0x100): RDMA descriptor transfer completed
		// RDAI (0x200): RDMA transfer aborted
		// OMBI (0x400): Outbound mailbox ready
		// CBRI (0x800): CSB bridge reset
		hisr = readl(&pciexp->bridge.pex_hisr);
		if (!hisr)
			break;

		hisr_handled |= hisr;

		PRINTK_DEBUG("%s:%d hisr: %08x", __FUNCTION__, __LINE__, hisr);

		// reset PCI Express Host Interrupt
		writel(hisr, &pciexp->bridge.pex_hisr);

		if (hisr & 0x380) {
			PRINTK_DEBUG("pex_rdma_stat: %08x", readl(&pciexp->bridge.pex_rdma_stat));

			// reset PCI Express Read DMA Status Register
			writel(0xffffffff, &pciexp->bridge.pex_rdma_stat);
		}

		if (hisr & 0x70) {
			PRINTK_DEBUG("pex_wdma_stat: %08x", readl(&pciexp->bridge.pex_wdma_stat));

			// reset PCI Express Write DMA Status Register
			writel(0xffffffff, &pciexp->bridge.pex_wdma_stat);
		}

		if (hisr & 0x400) {
			do {
				if (fix_reg_bits(devdata))
					wake_up_all(&devdata->io_wq);

				if (readl(devdata->host_read_wait)) {
					if (rfifo_len(devdata->dev2host) >= readl(devdata->host_read_wait)) {
						devdata->read_irqs++;
						writel(0, devdata->host_read_wait);
						wake_up_all(&devdata->read_wq);
					}
				}

				if (readl(devdata->host_write_wait)) {
					if (rfifo_unused(devdata->host2dev) >= readl(devdata->host_write_wait)) {
						devdata->write_irqs++;
						writel(0, devdata->host_write_wait);
						wake_up_all(&devdata->write_wq);
					}
				}

			} while(readl(&pciexp->bridge.pex_ombcr));
		}
	}

	return (hisr_handled) ? IRQ_HANDLED : IRQ_NONE;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
static irqreturn_t sgs_isr(int irq, void *dev_id, struct pt_regs *regs)
#else
static irqreturn_t sgs_isr(int irq, void *dev_id)
#endif
{
	struct devdata * devdata = (struct devdata *)dev_id;
	irqreturn_t ret;

	spin_lock(&devdata->lock);
	ret = _sgs_isr(devdata);
	spin_unlock(&devdata->lock);

	return ret;
}

struct sg_dev sgs_device = {
	.setup = sgs_setup,
	.reset = sgs_reset,
	.link = sgs_link,
	.unlink = sgs_unlink,
	.gen_inbound_mbox_irq = sgs_gen_inbound_mbox_irq,
	.isr = sgs_isr,
};

struct sg_dev sgs_device_old = {
	.setup = sgs_setup_old,
	.reset = sgs_reset_old,
	.gen_inbound_mbox_irq = sgs_gen_inbound_mbox_irq,
	.isr = sgs_isr,
};

